#topology {  
    border: 1px solid #000000;  
}

.node {
}

.node.fixed {
    fill: #C0C0C0;
}

.node text {
    font-size: 14px;
}

.link {
    stroke: #090909;
    stroke-opacity: .6;
    stroke-width: 2px;
}

.port circle {
    stroke: black;
    fill: #C5F9F9;
}

.port text {
    font-size: 10px;
}

